"""This module controls the directory structure used to save the results."""
import abc
import datetime
import json
import pathlib
from typing import Any, Callable, TypeVar

import pandas as pd  # pytype: disable=import-error
import petname  # pytype: disable=import-error
import pydantic  # pytype: disable=import-error

import cansig.types as types


class StructuredDir(abc.ABC):
    def __init__(self, path: types.Pathlike, create: bool = False) -> None:
        self.path = pathlib.Path(path)
        if create:
            self.create()

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.path})"

    def create(self) -> None:
        self.path.mkdir(parents=True, exist_ok=False)

    @property
    @abc.abstractmethod
    def valid(self) -> bool:
        pass


_Settings = TypeVar("_Settings", bound=pydantic.BaseModel)


def read_settings(factory: Callable[[Any], _Settings], path: types.Pathlike) -> _Settings:
    with open(path) as f:
        raw = json.load(f)
    # Annotating a callable which takes `**kwargs` is very tricky,
    # so we simply ignore the typecheck
    return factory(**raw)  # pytype: disable=wrong-arg-count


def save_settings(settings: pydantic.BaseModel, path: types.Pathlike) -> None:
    with open(path, "w") as f:
        f.write(settings.json())


def save_latent_representations(representations: pd.DataFrame, path: types.Pathlike) -> None:
    representations.to_csv(path, index=True, header=False)


def read_latent_representations(path: types.Pathlike) -> pd.DataFrame:
    return pd.read_csv(path, index_col=0, header=None)


def save_cluster_labels(labels, index, path: types.Pathlike) -> None:
    return pd.DataFrame(labels, index=index).to_csv(path, index=True, header=False)


def read_cluster_labels(path: types.Pathlike) -> pd.DataFrame:
    return pd.read_csv(path, index_col=0, header=False)


class IntegrationDir(StructuredDir):
    MODEL: str = "integration-settings.json"
    REPRESENTATIONS: str = "latent-representations.csv"

    def valid(self) -> bool:
        return self.path.is_dir() and self.latent_representations.is_file() and self.integration_settings.is_file()

    @property
    def latent_representations(self) -> pathlib.Path:
        return self.path / self.REPRESENTATIONS

    @property
    def integration_settings(self) -> pathlib.Path:
        return self.path / self.MODEL


class PostprocessingDir(StructuredDir):
    LABELS: str = "cluster-labels.csv"
    CLUSTER_SETTINGS: str = "cluster-settings.json"
    INTEGRATION_SETTINGS: str = IntegrationDir.MODEL
    GSEA_SETTINGS: str = "gsea-settings.json"

    def valid(self) -> bool:
        return (
            self.path.is_dir()
            and self.cluster_labels.is_file()
            and self.cluster_settings.is_file()
            and self.integration_settings.is_file()
        )

    @property
    def cluster_labels(self) -> pathlib.Path:
        return self.path / self.LABELS

    @property
    def cluster_settings(self) -> pathlib.Path:
        return self.path / self.CLUSTER_SETTINGS

    @property
    def integration_settings(self) -> pathlib.Path:
        return self.path / self.INTEGRATION_SETTINGS

    @property
    def gsea_settings(self) -> pathlib.Path:
        return self.path / self.GSEA_SETTINGS

    @property
    def gsea_output(self) -> pathlib.Path:
        # TODO(Pawel): Note, the desired output is still discussed.
        return self.path / "gsea-dataframe.csv"


def get_directory_name() -> str:
    """A string representing a unique name for the run."""
    now = datetime.datetime.now()  # current date and time
    date_time = now.strftime("%Y%m%d-%H%M%S")
    suffix = petname.generate(separator="-")
    return f"{date_time}-{suffix}"


def get_file(file_or_dir: types.Pathlike, ext: str) -> pathlib.Path:
    """If the specified file is a directory, checks if there is a unique
    file ending with `ext` inside it.

    Returns:
        `file_or_dir` if it is a regular file, otherwise the unique file
            with extension `ext` inside
        `ext`: suffix, e.g. ".csv", or ".txt"

    Raises:
        FileNotFoundError, if the file doesn't exist
        FileExistsError, if multiple files inside `file_or_dir` have
            matching extension `ext`
    """
    fod = pathlib.Path(file_or_dir)

    if not fod.exists():
        raise FileNotFoundError(f"File {file_or_dir} does not exist.")
    if fod.is_file():
        return fod
    elif fod.is_dir():
        candidates = list(fod.glob(f"*{ext}"))
        if len(candidates) == 0:
            raise FileNotFoundError(f"There are no files with extension {ext} in directory {fod}.")
        elif len(candidates) >= 2:
            raise FileExistsError(f"There are too many candidates in {fod}: {candidates}.")
        else:
            return candidates[0]
    else:
        raise ValueError(f"File {fod} is neither regular nor a directory.")



"""Gene expression analysis utilities."""
import pathlib
from typing import Iterable, Dict, List, Optional, Union
from typing import get_args  # pytype: disable=import-error
from typing import Literal  # pytype: disable=not-supported-yet

import anndata  # pytype: disable=import-error
import gseapy as gp  # pytype: disable=import-error
import numpy as np
import pandas as pd  # pytype: disable=import-error
import pydantic  # pytype: disable=import-error
import scanpy as sc  # pytype: disable=import-error
from scanpy.tools._rank_genes_groups import _Method  # pytype: disable=import-error

# This is kind of ugly because the defaults are not obvious but since it can be a path
# and path are string we can't really differentiate them.
_GENESETS = Union[str, pathlib.Path]


def genes_to_string(genes: List[str], sep: str = ";") -> str:
    """Joins a list of strings"""
    return sep.join(genes)


class GeneExpressionAnalysis:
    """Interface to perform the gene expression analysis for Step 2 of the pipeline.

    Performs first differential gene expression analysis and then GSEA.

    Attrs:
        cluster_name: name of the column in which to find the cluster labels
            in the AnnData object
        group_names: clusters to perform the analysis for, default is all,
            else iterable of valid names of the clusters
        method: name of the method to use for differential analysis,
            see "Bias, robustness and scalability in single-cell differential expression
            analysis" Soneson & Robinson for more info on the choice
        n_diff_genes: number of differentially expressed genes for every cluster
            to keep for further analysis
        gene_sets: gene set to perform the GSEA on. Can be a string if valid
            in the ENRICHR API.
            Otherwise, it can be a path to the .gmt file downloaded from the MSigDB site
        permutation_num: number of permutations computed in GSEA.
            Lower reduces the computational complexity,
            higher leads to more robust results.
    """

    def __init__(
        self,
        cluster_name: str = "cluster",
        group_names: Union[Iterable[str], Literal["all"]] = "all",
        method: _Method = "wilcoxon",
        n_diff_genes: Optional[int] = None,
        gene_sets: _GENESETS = "MSigDB_Hallmark_2020",
        permutation_num: int = 500,
    ) -> None:

        self.cluster_name = cluster_name
        self.group_names = group_names
        self.method = method
        self._check_method()
        self.n_diff_genes = n_diff_genes
        self.gene_sets = gene_sets
        self.permutation_num = permutation_num

    def _check_method(self) -> None:
        """Verifies that the method called is valid"""
        if self.method is not None:
            allowed = get_args(get_args(_Method)[0])
            if self.method not in allowed:
                raise ValueError(f"The method must be one of {allowed}.")

    def diff_gex(self, adata: anndata.AnnData) -> Dict[str, pd.DataFrame]:
        """Computes the differential gene expression between cluster groups

        Attrs:
            adata: AnnData object containing the clustering labels and gene counts in X

        Returns:
            adata_copy: copy of the AnnData object with
                log-normalized and scaled gene expression
                and performed gene expression analysis
        """

        if self.n_diff_genes is None:
            self.n_diff_genes = adata.shape[1]

        if self.n_diff_genes > adata.shape[1]:
            raise ValueError(f"n_diff_genes of {self.n_diff_genes} is greater" f"then the number of genes in adata.")

        adata_copy = adata.copy()
        sc.pp.normalize_total(adata_copy)
        sc.pp.log1p(adata_copy)

        sc.tl.rank_genes_groups(
            adata_copy,
            groupby=self.cluster_name,
            groups=self.group_names,
            method=self.method,
        )
        ranked_genes = adata_copy.uns["rank_genes_groups"]

        gene_order = {}
        for label in ranked_genes["names"].dtype.names:
            gene_order[label] = pd.DataFrame(
                ranked_genes["scores"][label][: self.n_diff_genes],
                index=ranked_genes["names"][label][: self.n_diff_genes],
                columns=["logfold"],
            )

        return gene_order

    def perform_gsea(self, diff_genes: Dict[str, pd.DataFrame]) -> pd.DataFrame:
        """Performs the gene set enrichment analysis for every cluster group using pre-ranked
        differentially expressed genes.

        Attrs:
            adata: AnnData object containing the clustering labels and
            gene counts in X
            diff_genes: dictionary with cluster label as key and a
            pd.DataFrame containing the
            n_diff_genes most overexpressed genes ranked by logfold
            change in the corresponding cluster

        Returns:
            pathways: dictionary with the cluster name as
            key and the pd.DataFrame with
            the 5 most enriched pathways found by GSEA as value

        Note:
            the columns of the pathways[label] pd.DataFrames
            are the classical GSEA
            columns ie
                - 'es' (enrichment score),
                - 'nes' (normalized enrichemnt score)
                - 'pval' (p value)
                - 'fdr' (FDR Benjamini Hochberg corrected p value)
                - 'geneset_size' (size of the geneset of the term)
                - 'matched_size' (size of the genes of the geneset
                found in those studied)
                - 'genes' (list of genes matched)
                - 'ledge_genes' (leading edge genes - see GSEA doc)
            Additionally we add the column
                - 'genes_for_scoring' (list of the genes that are
                    positively ranked in our
                    dataset and matched to the term that we will
                    use for scoring)
        """
        gsea_dfs = []
        for label, gene_rank in diff_genes.items():

            print(f"GSEA for cluster {label}")
            gs_res = gp.prerank(
                rnk=gene_rank,
                gene_sets=self.gene_sets,
                processes=4,
                permutation_num=self.permutation_num,  # reduce number for speed
                outdir=None,
                format="png",
                seed=6,
            )

            gsea_df = gs_res.res2d.sort_values(by="nes", ascending=False)
            pos_genes = gs_res.ranking[gs_res.ranking.sort_values() >= 0].index

            genes_for_scoring = []

            for i in range(gsea_df.shape[0]):
                genes = np.intersect1d(gsea_df.iloc[i].genes.split(";"), pos_genes)
                genes_scoring = genes_to_string(genes)
                genes_for_scoring.append(genes_scoring)

            genes_for_scoring = np.array(genes_for_scoring)
            gsea_df["genes_for_scoring"] = genes_for_scoring
            gsea_df["cluster"] = label
            gsea_dfs.append(gsea_df)

        return pd.concat(gsea_dfs)


class GeneExpressionConfig(pydantic.BaseModel):
    method: _Method = pydantic.Field(default="t-test")
    gene_sets: _GENESETS = pydantic.Field(default="MSigDB_Hallmark_2020")
    n_diff_genes: Optional[int] = pydantic.Field(default=None)
    permutation_num: int = pydantic.Field(default=100)


def gex_factory(cluster_name: str, config: GeneExpressionConfig) -> GeneExpressionAnalysis:
    return GeneExpressionAnalysis(
        cluster_name=cluster_name,
        group_names="all",
        method=config.method,
        n_diff_genes=config.n_diff_genes,
        gene_sets=config.gene_sets,
        permutation_num=config.permutation_num,
    )



"""Convenient generic type definitions (used in multiple places)."""
import pathlib
from typing import cast, NewType, Sequence, Union

import numpy as np
import pandas as pd

Pathlike = Union[pathlib.Path, str]  # Represents input paths

# Represents latent codes. Index should identify cells.
Latents = NewType("Latents", pd.DataFrame)


def create_latents(latent_codes: np.ndarray, index: Sequence) -> Latents:
    """A factory method for `Latents`.

    Args:
        latent_codes: numpy array, shape (n_cells, k),
            where k is the latent space dimension
        index: cell identifiers

    Returns:
        Latents
    """
    df = pd.DataFrame(latent_codes, index=index)
    return cast(Latents, df)


def write_latents(latent_codes: Latents, path: Pathlike) -> None:
    """Writes latent codes to the disk.

    Args:
        latent_codes: latent codes
        path: path where the latent codes should be dumped

    See Also:
        read_latents, used to read latent codes from the disk
    """
    latent_codes = cast(pd.DataFrame, latent_codes)
    latent_codes.to_csv(path)


def read_latents(path: Pathlike) -> Latents:
    """Reads latent codes from the disk.

    Args:
        path: path where

    Returns:
        latent codes

    See Also:
        write_latents, function used to dump the latent codes to the disk
    """
    codes = pd.read_csv(path, index_col=0)
    return cast(Latents, codes)



"""Everything needed for the tutorial."""
import dataclasses
import hashlib
import io
import pathlib
import pickle
import requests
import zipfile
from typing import Dict, List, Tuple, Union

# TODO(Pawel): Remove this comment when pytype starts supporting Literal
from typing import Literal  # pytype: disable=not-supported-yet

# TODO(Pawel): Remove these comments when anndata and pandas work with pytype
import anndata as ad  # pytype: disable=import-error
import pandas as pd  # pytype: disable=import-error

# TODO(Florian, Pawel): Find a better place to host the data.
_DATA_URL = "https://polybox.ethz.ch/index.php/s/59uBkhs0JmKQGcJ/download"
_PREPROCESSED_URL = ""


@dataclasses.dataclass
class Dataset:
    url: str
    md5sum: str
    filepath: str


_RAW_DATA = Dataset(url=_DATA_URL, md5sum="3489aaa0f4cc15c61b4b910110306d11", filepath="data/raw-data.zip")


def _file_ok(path: pathlib.Path, md5sum: str, open_mode: Literal["r", "rb"] = "rb") -> bool:
    """Checks if `path` exists and has the desired MD5 checksum."""
    if not path.is_file():
        return False

    with open(path, open_mode) as f:
        checksum = hashlib.md5(f.read()).hexdigest()

    return checksum == md5sum


def _download_simple(
    open_file,
    source_url: str,
    verbose: bool,
    chunk_size: int,
) -> None:
    response = requests.get(source_url, stream=True)
    total = response.headers.get("content-length")
    if total is None:
        open_file.write(response.content)
    else:
        progress_bar_len = 30
        total = int(total)
        ready = 0
        for chunk in response.iter_content(chunk_size=chunk_size):
            open_file.write(chunk)
            ready += len(chunk)

            progress = int(progress_bar_len * ready / total)
            progress_str = "#" * progress + (progress_bar_len - progress) * "."

            if verbose:
                print(f"\r[{progress_str}]", end="")


def download_file_to_path(
    source_url: str,
    path: Union[str, pathlib.Path],
    md5sum: str,
    verbose: bool = True,
    _chunk_size: int = 8192,
) -> None:
    """This function downloads data from `source_url`
    and saves them to `path`. Compares the MD5 checksum
    with the provided one, to make sure that the file has
    been downloaded properly

    Args:
        source_url: the URL with the file to be downloaded
        path: where the file should be downloaded to. If the file
         already exists there, it will just be read
        md5sum: desired MD5 checksum
        verbose: whether to print our progress
        _chunk_size: controls the size of each data chunk during download
    """
    open_mode = "rb"
    path = pathlib.Path(path)

    # Make sure the parent directory exists
    path.parent.mkdir(parents=True, exist_ok=True)

    # If file exists, do nothing
    if _file_ok(path=path, md5sum=md5sum, open_mode=open_mode):
        if verbose:
            print(f"The file already exists at {path}.")
        return

    # Otherwise, download it
    if verbose:
        print(f"Retrieving the file from {source_url}...")
    with open(path, "wb") as f:
        _download_simple(open_file=f, source_url=source_url, chunk_size=_chunk_size, verbose=verbose)

    if not _file_ok(path=path, md5sum=md5sum, open_mode=open_mode):
        raise IOError("The downloaded file is corrupted.")

    if verbose:
        print(f"\nFile downloaded to {path}.")


def load_data(preprocessed: bool = False) -> Tuple[List[ad.AnnData], pd.DataFrame, List[Dict]]:
    """
    load_data returns the data needed for the tutorial. See the preprocessing tutorial
    for details.

    Args:
        preprocessed (bool): set to True to download the already preprocessed data
        (default: False).

    Returns: Tuple containing a list of AnnData objects for the raw data, gene
    annotations and scoring list.
    """
    print("Downloading tutorial data. This might take a while.")

    if preprocessed:
        responds = requests.get(_PREPROCESSED_URL)
    else:
        responds = requests.get(_DATA_URL)

    if responds.status_code != 200:
        print("Connection error, please try again later. If the problem remains please raise an issue on github.")
        responds.raise_for_status()
    # TODO: Add progressbar to the download.
    content = responds.content
    buffer = zipfile.ZipFile(io.BytesIO(content)).read("data.pickle")
    data = pickle.load(io.BytesIO(buffer))
    return data


if __name__ == "__main__":
    download_file_to_path(source_url=_RAW_DATA.url, md5sum=_RAW_DATA.md5sum, path=_RAW_DATA.filepath, verbose=True)



"""A dummy setup file. It may not be required in the future.

See the warning at
https://setuptools.pypa.io/en/latest/setuptools.html
"""
import setuptools


if __name__ == "__main__":
    setuptools.setup()
